/*
	basicConsumer - Tutorial 4 - Retrieving level 1 content and decoding the results

	The goal of this tutorial is to extend Tutorial 3 to request and retrieve level 1 (MarketPrice) data from the OMM-based data
	Provider.  To support the level1 content, we offload the decoding of messages into separate files to allow
	for easy management and support for additional functionality.

	The series includes the following:

		o Tutorial 1 - Creating a basic barebones ETA consumer application
		o Tutorial 2 - Establishing a connection to an OMM-based data Provider
		o Tutorial 3 - Establishing a session with an OMM-based data Provider
		o Tutorial 4 - Retrieving level 1 content and decoding the results
		o Tutorial 5 - Retrieving level 2 content and decoding the results
 */

/* System headers */
#include <stdlib.h>

/* Application headers */
#include "basicConsumer.h"
#include "basicDirectoryHandler.h"
#include "basicDictionaryHandler.h"
#include "basicMsgHandler.h"

/* ETA headers */
#include "rtr/rsslReactor.h"

/* Server/Provider connection details */
//Remove for RTO V2
//#define RTDS_SERVER_NAME	"elektron"
//#define RTDS_SERVER_PORT	"14002"

//RTO V2
char clientId[] = "<Client ID>";
char clientSecret[] = "<Client Secret>";
char location[] = "ap-northeast-1";



/* Name of our market data service, item and domain of our subscription */
#define SERVICE	"ELEKTRON_DD"
char itemName[] = "/LSEG.L";
RsslUInt8 domainType = RSSL_DMT_MARKET_PRICE;

/* RDM objects used at startup */
RsslReactorConnectOptions connectionOptions;
RsslRDMLoginRequest loginRequest;				// Login Object
RsslRDMDirectoryRequest directoryRequest;		// Directory Object
BasicService service = { SERVICE, 0 };			// Service Object

/* RsslReactor components*/
fd_set readFds, exceptFds;
RsslReactor *pReactor = NULL;

/* Local functions */
void init();
void mainEventLoop();
RsslReactorCallbackRet channelEventCallback(RsslReactor *pReactor, RsslReactorChannel *pReactorChannel, RsslReactorChannelEvent *pConnEvent);
RsslReactorCallbackRet loginMsgCallback(RsslReactor *pReactor, RsslReactorChannel *pChannel, RsslRDMLoginMsgEvent *pLoginMsgEvent);
RsslReactorCallbackRet directoryMsgCallback(RsslReactor *pReactor, RsslReactorChannel *pChannel, RsslRDMDirectoryMsgEvent *pDirectoryMsgEvent);
RsslReactorCallbackRet dictionaryMsgCallback(RsslReactor *pReactor, RsslReactorChannel *pReactorChannel, RsslRDMDictionaryMsgEvent *pDictionaryMsgEvent);
RsslReactorCallbackRet defaultMsgCallback(RsslReactor *pReactor, RsslReactorChannel *pChannel, RsslMsgEvent* pMsgEvent);
void closeConnection(RsslReactor *pReactor, RsslReactorChannel *pChannel);
void cleanUpAndExit(int code);
void exitApp(int code);

/*** MAIN ***/
int main(int argc, char **argv)
{
	// Initialization
	init();

	// Main Event Loop
	mainEventLoop();

	// Exit
	cleanUpAndExit(0);
}

/* Initialize RSSL and setup the main communication protocol details as outlined in the 3 steps within the introduction */
void init()
{
	RsslError error;

	RsslReactorOMMConsumerRole consumerRole;		// Channel Consumer role - client (consumer) connecting into a server (provider)
	RsslCreateReactorOptions reactorOpts;
	RsslErrorInfo rsslErrorInfo;
	//RTO V2
	RsslReactorOAuthCredential oauthCredential;
	
	
	RsslBuffer clientIdBuffer = RSSL_INIT_BUFFER;
	RsslBuffer clientSecretBuffer = RSSL_INIT_BUFFER;
	RsslBuffer locationBuffer = RSSL_INIT_BUFFER;

	RsslReactorConnectInfo cInfo;

	rsslClearReactorOAuthCredential(&oauthCredential);
	rsslClearReactorConnectInfo(&cInfo);

	clientIdBuffer.data = clientId;
	clientIdBuffer.length = strlen(clientId);

	clientSecretBuffer.data = clientSecret;
	clientSecretBuffer.length = strlen(clientSecret);

	locationBuffer.data = location;
	locationBuffer.length = strlen(location);



	

	/* Initialize RSSL. The locking mode RSSL_LOCK_GLOBAL_AND_CHANNEL is required to use the RsslReactor. */
	/* RsslReactor internally utilizes multiple threads */
	if (rsslInitialize(RSSL_LOCK_GLOBAL_AND_CHANNEL, &error) != RSSL_RET_SUCCESS)
	{
		printf("rsslInitialize(): failed <%s>\n", error.text);
		exit(RSSL_RET_FAILURE);
	}

	/* After this consumer successfully connects into a server (RTDS), we must log in.*/
	/* Initialize a default login request (Use 1 as the Login Stream ID). */
	if (rsslInitDefaultRDMLoginRequest(&loginRequest, 1) != RSSL_RET_SUCCESS)
	{
		printf("rsslInitDefaultRDMLoginRequest() failed\n");
		cleanUpAndExit(-1);
	}

	/* Initialize the default directory request (Use 2 as the Directory Stream Id) */
	if (rsslInitDefaultRDMDirectoryRequest(&directoryRequest, 2) != RSSL_RET_SUCCESS)
	{
		printf("rsslInitDefaultRDMDirectoryRequest() failed\n");
		cleanUpAndExit(-1);
	}

	rsslClearOMMConsumerRole(&consumerRole);

	/* Set the messages to send when the channel is up */
	consumerRole.pLoginRequest = &loginRequest;
	consumerRole.pDirectoryRequest = &directoryRequest;
	
	/**************************************************************************
	 * Setup callback functions for our connection channel - consumer role 
	 **************************************************************************/

	/* Login callback */
	consumerRole.loginMsgCallback = loginMsgCallback;

	/* Directory callback */
	consumerRole.directoryMsgCallback = directoryMsgCallback;

	/* Connection/Channel callback */
	consumerRole.base.channelEventCallback = channelEventCallback;

	/* Market Data Message callback */
	consumerRole.base.defaultMsgCallback = defaultMsgCallback;

	/* Load Dictionaries - local or server */
	consumerRole.dictionaryMsgCallback = loadLocalDictionary() ? NULL : dictionaryMsgCallback;
	consumerRole.dictionaryDownloadMode = (consumerRole.dictionaryMsgCallback ? RSSL_RC_DICTIONARY_DOWNLOAD_FIRST_AVAILABLE :
																				RSSL_RC_DICTIONARY_DOWNLOAD_NONE);
	//RTO V2
	oauthCredential.clientId = clientIdBuffer;
	oauthCredential.clientSecret = clientSecretBuffer;

	consumerRole.pOAuthCredential = &oauthCredential;


	/* Create our rsslReactor used to manage the ETA channel */
	rsslClearCreateReactorOptions(&reactorOpts);
	if (!(pReactor = rsslCreateReactor(&reactorOpts, &rsslErrorInfo)))
	{
		printf("Error: %s", rsslErrorInfo.rsslError.text);
		exit(-1);
	}

	// Prepare our connection properties
	rsslClearReactorConnectOptions(&connectionOptions);

	//connectionOptions.rsslConnectOptions.connectionInfo.unified.address = RTDS_SERVER_NAME;
	//connectionOptions.rsslConnectOptions.connectionInfo.unified.serviceName = RTDS_SERVER_PORT;
	//connectionOptions.rsslConnectOptions.guaranteedOutputBuffers = 500;
	//connectionOptions.rsslConnectOptions.majorVersion = RSSL_RWF_MAJOR_VERSION;
	//connectionOptions.rsslConnectOptions.minorVersion = RSSL_RWF_MINOR_VERSION;

	//RTO V2
	cInfo.rsslConnectOptions.connectionType = RSSL_CONN_TYPE_ENCRYPTED;
	cInfo.rsslConnectOptions.encryptionOpts.encryptedProtocol = RSSL_CONN_TYPE_SOCKET;
	cInfo.rsslConnectOptions.guaranteedOutputBuffers = 500;
	cInfo.rsslConnectOptions.majorVersion = RSSL_RWF_MAJOR_VERSION;
	cInfo.rsslConnectOptions.minorVersion = RSSL_RWF_MINOR_VERSION;
	cInfo.location = locationBuffer;
	cInfo.enableSessionManagement = RSSL_TRUE;
	connectionOptions.reactorConnectionList = &cInfo;
	connectionOptions.connectionCount = 1;
	connectionOptions.reconnectAttemptLimit = -1;	// Keep trying to reconnect upon failure
	connectionOptions.reconnectMinDelay = 5000;		// Reconnect interval (milliseconds)
	connectionOptions.reconnectMaxDelay = 10000;
	
	/* Reactor fd monitoring */
	FD_ZERO(&readFds);
	FD_ZERO(&exceptFds);

	/* Set the reactor's event file descriptor on our descriptor set. This, along with the file descriptors
	 * of RsslReactorChannels, will notify us when we should call rsslReactorDispatch(). */
	FD_SET(pReactor->eventFd, &readFds);

	/* Initialize connection */
	if (rsslReactorConnect(pReactor, &connectionOptions, (RsslReactorChannelRole*)&consumerRole, &rsslErrorInfo) != RSSL_RET_SUCCESS)
	{
		printf("Error rsslReactorConnect(): %s\n", rsslErrorInfo.rsslError.text);
		exit(-1);
	}
}

void mainEventLoop()
{
	RsslReactorDispatchOptions dispatchOpts;
	RsslErrorInfo rsslErrorInfo;
	int ret;

	rsslClearReactorDispatchOptions(&dispatchOpts);

	/* Main loop - select() to trigger event notification then dispatch (rsslReactorDispatch()) */
	do
	{
		fd_set useReadFds = readFds, useExceptFds = exceptFds;

		/* detect channel notifications */
		ret = select(FD_SETSIZE, &useReadFds, NULL, &useExceptFds, NULL);

		if (ret < 0)
		{
#ifdef _WIN32
                        if (WSAGetLastError() == WSAEINTR)
                                continue;

                        printf("Error: select: %d\n", WSAGetLastError());
#else
                        if (errno == EINTR)
                                continue;

                        perror("select");
#endif
                        cleanUpAndExit(-1);
		}

		/* Call rsslReactorDispatch().  This will handle any events that have occurred on our rssl channel and
		 * dispatched via the callbacks specified for our consumerRole as referenced within init().
		   A return value of RSSL_RET_SUCCESS indicates there may be more to process */
		while ((ret = rsslReactorDispatch(pReactor, &dispatchOpts, &rsslErrorInfo)) > RSSL_RET_SUCCESS)
			;

		if (ret < RSSL_RET_SUCCESS)
			printf("rsslReactorDispatch() failed: %s\n", rsslErrorInfo.rsslError.text);

	} while (ret >= RSSL_RET_SUCCESS);
}

/*
* Processes events about the state of an RsslReactorChannel.
*/
RsslReactorCallbackRet channelEventCallback(RsslReactor *pReactor, RsslReactorChannel *pReactorChannel, RsslReactorChannelEvent *pConnEvent)
{
	switch (pConnEvent->channelEventType)
	{
	case RSSL_RC_CET_CHANNEL_UP:
	{
		/* A channel that we have requested via rsslReactorConnect() has come up.  We must now
		 * associated the new channel descriptor within our select() watchlist. This will drive 
		 * the process of detecting message events, i.e. Login, Directory, Dictionary etc. */
		printf("Connection up! Channel fd=%lld\n\n", pReactorChannel->socketId);

		/* Set file descriptor. */
		FD_SET(pReactorChannel->socketId, &readFds);
		FD_SET(pReactorChannel->socketId, &exceptFds);

		/* Note: Refer to the complete rsslConsumer example to set channel properites appropriate for your environment */
		break;
	}
	case RSSL_RC_CET_CHANNEL_READY:
		printf("Channel is ready.  Subscribing to %s\n", itemName);
		sendItemRequest(pReactor, pReactorChannel, itemName, service.serviceId, domainType);
		break;

	case RSSL_RC_CET_FD_CHANGE:
	{
		/* The file descriptor representing the RsslReactorChannel has been changed.
		* Update our file descriptor sets. */
		printf("Fd change: %lld to %lld\n", pReactorChannel->oldSocketId, pReactorChannel->socketId);
		FD_CLR(pReactorChannel->oldSocketId, &readFds);
		FD_CLR(pReactorChannel->oldSocketId, &exceptFds);
		FD_SET(pReactorChannel->socketId, &readFds);
		FD_SET(pReactorChannel->socketId, &exceptFds);
		break;
	}
	case RSSL_RC_CET_CHANNEL_DOWN:
	{
		/* The channel has failed and has gone down.  Print the error, close the channel, and reconnect later. */
		printf("Connection down: Channel fd=%lld.\n", pReactorChannel->socketId);

		if (pConnEvent->pError)
			printf("	Error text: %s\n\n", pConnEvent->pError->rsslError.text);

		closeConnection(pReactor, pReactorChannel);
		break;
	}
	case RSSL_RC_CET_CHANNEL_DOWN_RECONNECTING:
	{
		printf("Connection down, reconnecting.  Channel fd=%lld\n", pReactorChannel->socketId);

		if (pReactorChannel->socketId != REACTOR_INVALID_SOCKET)
		{
			FD_CLR(pReactorChannel->socketId, &readFds);
			FD_CLR(pReactorChannel->socketId, &exceptFds);
		}
		printf("	Error text: %s\n\n", pConnEvent->pError->rsslError.text);
		break;
	}
	case RSSL_RC_CET_WARNING:
	{
		/* We have received a warning event for this channel. Print the information and continue. */
		printf("Received warning for Channel fd=%lld.\n", pReactorChannel->socketId);
		printf("	Error text: %s\n\n", pConnEvent->pError->rsslError.text);
		break;
	}
	default:
	{
		printf("Unknown connection event!\n");
		cleanUpAndExit(-1);
	}
	}

	return RSSL_RC_CRET_SUCCESS;
}

/*
* Processes the information contained in Login responses.
* Copies the refresh so we can use it to know what features are
* supported(for example, whether posting is supported
* by the provider).
*/
RsslReactorCallbackRet loginMsgCallback(RsslReactor *pReactor, RsslReactorChannel *pChannel, RsslRDMLoginMsgEvent *pLoginMsgEvent)
{
	RsslBuffer tempBuffer = { 1024, (char*)alloca(1024) };
	RsslRDMLoginMsg *pLoginMsg = pLoginMsgEvent->pRDMLoginMsg;

	// Verify login event
	if (!pLoginMsg)
	{
		RsslErrorInfo *pError = pLoginMsgEvent->baseMsgEvent.pErrorInfo;
		printf("loginMsgCallback: %s(%s)\n", pError->rsslError.text, pError->errorLocation);
		closeConnection(pReactor, pChannel);
		return RSSL_RC_CRET_SUCCESS;
	}

	// Determine login type
	switch (pLoginMsg->rdmMsgBase.rdmMsgType)
	{
	case RDM_LG_MT_REFRESH:
	{
		RsslRDMLoginRefresh *pLoginRefresh = &pLoginMsg->refresh;

		/* get state information */
		rsslStateToString(&tempBuffer, &pLoginRefresh->state);
		printf("\nReceived Login response: %.*s\n\n", tempBuffer.length, tempBuffer.data);

		break;
	}
	case RDM_LG_MT_STATUS:
		printf("\nReceived Login StatusMsg\n");
		if (pLoginMsg->status.flags & RDM_LG_STF_HAS_STATE)
		{
			/* get state information */
			rsslStateToString(&tempBuffer, &pLoginMsg->status.state);
			printf("	%.*s\n\n", tempBuffer.length, tempBuffer.data);
		}
		break;

	default:
		printf("\nReceived Unhandled Login Msg Class: %d\n", pLoginMsg->rdmMsgBase.rdmMsgType);
		break;
	}

	return RSSL_RC_CRET_SUCCESS;
}

/*
* Processes information contained in Directory responses.
* Searches the refresh for the service the application
* wants to use for this connection and stores the information
* so we can request items from that service.
*/
RsslReactorCallbackRet directoryMsgCallback(RsslReactor *pReactor, RsslReactorChannel *pChannel, 
											RsslRDMDirectoryMsgEvent *pDirectoryMsgEvent)
{
	RsslRDMDirectoryMsg *pDirectoryMsg = pDirectoryMsgEvent->pRDMDirectoryMsg;

	if (!pDirectoryMsg)
	{
		RsslErrorInfo *pError = pDirectoryMsgEvent->baseMsgEvent.pErrorInfo;
		printf("processDirectoryResponse: %s(%s)\n", pError->rsslError.text, pError->errorLocation);
		closeConnection(pReactor, pChannel);
		return RSSL_RC_CRET_FAILURE;
	}

	/* Process the directory response */
	return(processDirectoryResponse(pDirectoryMsgEvent, &service));
}

/*
* Processes information contained in Dictionary responses.
* Takes the payload of the messages and caches them
* using the RSSL utilities for decoding dictionary info.
*/
RsslReactorCallbackRet dictionaryMsgCallback(RsslReactor *pReactor, RsslReactorChannel *pReactorChannel, 
											 RsslRDMDictionaryMsgEvent *pDictionaryMsgEvent)
{
	char	errTxt[256];
	RsslBuffer errorText = { 255, (char*)errTxt };
	RsslReactorCallbackRet ret = RSSL_RC_CRET_SUCCESS;

	RsslRDMDictionaryMsg *pDictionaryMsg = pDictionaryMsgEvent->pRDMDictionaryMsg;

	if (!pDictionaryMsg)
	{
		RsslErrorInfo *pError = pDictionaryMsgEvent->baseMsgEvent.pErrorInfo;
		printf("dictionaryResponseCallback: %s(%s)\n", pError->rsslError.text, pError->errorLocation);
		ret = RSSL_RC_CRET_FAILURE;
	}
	else
		ret = processDictionaryResponse(pReactor, pReactorChannel, pDictionaryMsgEvent);

	if (ret == RSSL_RC_CRET_FAILURE)
		closeConnection(pReactor, pReactorChannel);

	return(ret);
}

/*
* Processes all RSSL messages that aren't processed by any domain-specific callback functions.  Responses for
* items requested by the function are handled here.
*
* Returns: RSSL_RC_CRET_SUCCESS (success), RSSL_RC_CRET_FAILURE (otherwise)
*/
RsslReactorCallbackRet defaultMsgCallback(RsslReactor *pReactor, RsslReactorChannel *pChannel, RsslMsgEvent* pMsgEvent)
{
	RsslReactorCallbackRet ret = generalMsgCallback(pReactor, pChannel, pMsgEvent);

	if (ret == RSSL_RC_CRET_FAILURE)
	{
		/* Because this application is single - threaded, it is safe to call it inside callback functions. */
		closeConnection(pReactor, pChannel);
	}

	return(ret);
}

/*
* Closes a channel in the RsslReactor
*/
void closeConnection(RsslReactor *pReactor, RsslReactorChannel *pChannel)
{
	RsslErrorInfo rsslErrorInfo;

	if (pChannel->socketId != REACTOR_INVALID_SOCKET)
	{
		FD_CLR(pChannel->socketId, &readFds);
		FD_CLR(pChannel->socketId, &exceptFds);
	}

	if (rsslReactorCloseChannel(pReactor, pChannel, &rsslErrorInfo) != RSSL_RET_SUCCESS)
	{
		printf("rsslReactorCloseChannel() failed: %s\n", rsslErrorInfo.rsslError.text);
	}
}


/*
 * Cleans up and exits the application.
 */
void cleanUpAndExit(int code)
{
	RsslErrorInfo rsslErrorInfo;

	if (pReactor && rsslDestroyReactor(pReactor, &rsslErrorInfo) != RSSL_RET_SUCCESS)
	{
		printf("Error cleaning up reactor: %s\n", rsslErrorInfo.rsslError.text);
		exitApp(-1);
	}

	rsslUninitialize();

	exitApp(code);
}

void exitApp(int code)
{
	printf("Exiting.\n");
	exit(code);
}


